// Scintilla source code edit control
/** @file LexHyper.cxx
 ** Lexer for Hyper.
 **
 ** Copyright (C) 2006, 2007 by Kevin André <hyperquantum@gmail.com>
 **/

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <ctype.h>
#include <stdarg.h>

#include "Platform.h"

#include "PropSet.h"
#include "Accessor.h"
#include "StyleContext.h"
#include "KeyWords.h"
#include "Scintilla.h"
#include "SciLexer.h"


static inline bool isAWordChar(int ch) {
	return isascii(ch) && (isalnum(ch) || ch == '_');
}

static inline bool isAWordStart(int ch) {
	return isascii(ch) && isalpha(ch);
}

static inline bool isOperator(int ch) {
	switch (ch) {
	case '+': case '-': case '*': case '/': case '\\':
	case '!': case '$': case '%': case '&': case '?':
	case '.': case ',': case ';': case ':':
	case '(': case ')': case '[': case ']': case '{': case '}':
	case '<': case '>':
		return true;
	default:
		return false;
	}
}

static bool IsSpaceEquiv(int state) {
	return (state == SCE_HYPER_COMMENT_BLOCK)
				|| (state == SCE_HYPER_EMB_COMMENT_LINE);
}


static void ColouriseHyperDoc(unsigned int startPos, int length, int initStyle,
                           WordList *keywordlists[], Accessor &styler)
{
	
	StyleContext sc(startPos, length, initStyle, styler);
	WordList &keywords = *keywordlists[0];
	
	int visibleChars = 0;
	int curLine = styler.GetLine(startPos);
    int blockCommentDepth =
		(curLine > 0) ? styler.GetLineState(curLine - 1) : 0;
	
	for (; sc.More(); sc.Forward()) {
		if (sc.atLineStart) {
			//Platform::DebugPrintf("Seeing new line\n");
			visibleChars = 0;
			curLine = styler.GetLine(sc.currentPos);
            styler.SetLineState(curLine, blockCommentDepth);
		}
		
		// Determine if the current state should terminate.
		switch (sc.state) {
			case SCE_HYPER_EMB_COMMENT_LINE:
				if (sc.atLineStart) {
					sc.SetState(SCE_HYPER_DEFAULT);
				}
				else if (sc.Match('#', '#')) {
					// comment will continue until end of line
					sc.SetState(SCE_HYPER_EOL_COMMENT_LINE);
				}
				else if (sc.Match('#')) {
					sc.ForwardSetState(SCE_HYPER_DEFAULT);
				}
				break;
			case SCE_HYPER_EOL_COMMENT_LINE:
				if (sc.atLineStart) {
					sc.SetState(SCE_HYPER_DEFAULT);
				}
				break;
			case SCE_HYPER_COMMENT_BLOCK:
				// end of block comment ?
				if (visibleChars == 0 && sc.Match('\\', '#')
					&& blockCommentDepth > 0)
				{
					//Platform::DebugPrintf("End of block found\n");
					blockCommentDepth--;
					//Platform::DebugPrintf(" Level down to %i\n", blockCommentDepth);
                    curLine = styler.GetLine(sc.currentPos);
                    styler.SetLineState(curLine, blockCommentDepth);
                    sc.Forward();
				}
				else if (visibleChars == 0 && sc.Match('/', '#')) {
					//Platform::DebugPrintf("Nested block start found\n");
					blockCommentDepth++;
					//Platform::DebugPrintf(" Level up to %i\n", blockCommentDepth);
					curLine = styler.GetLine(sc.currentPos);
					styler.SetLineState(curLine, blockCommentDepth);
					sc.Forward();
				}
				
				if (sc.atLineEnd && blockCommentDepth == 0) {
					//Platform::DebugPrintf("Comment block mode stopped\n");
					sc.SetState(SCE_HYPER_DEFAULT);
				}
				break;
			case SCE_HYPER_KEYWORD:
				// UNREACHABLE
				break;
			case SCE_HYPER_NUMBER:
				if (!isAWordChar(sc.ch)) {
					sc.SetState(SCE_HYPER_DEFAULT);
				}
				break;
			case SCE_HYPER_STRING:
				if (sc.atLineEnd) { // string terminated by end-of-line (not allowed in Hyper)
					// TODO
					//sc.SetState(SCE_HYPER_DEFAULT);
				} else if (sc.ch == '\"') {
					sc.ForwardSetState(SCE_HYPER_DEFAULT);
				}
				break;
			case SCE_HYPER_CHAR:
				if (sc.atLineEnd) { // char terminated by end-of-line (not allowed in Hyper)
					// TODO
					//sc.SetState(SCE_HYPER_DEFAULT);
				} else if (sc.ch == '\'') {
					sc.ForwardSetState(SCE_HYPER_DEFAULT);
				}
				break;
			case SCE_HYPER_OPERATOR:
				sc.SetState(SCE_HYPER_DEFAULT);
				break;
			case SCE_HYPER_IDENTIFIER:
				if (!isAWordChar(sc.ch)) {
					char s[1000];
					sc.GetCurrent(s, sizeof(s));
					if (keywords.InList(s)) {
						sc.ChangeState(SCE_HYPER_KEYWORD);
					}
					sc.SetState(SCE_HYPER_DEFAULT);
				}
				break;
			case SCE_HYPER_ERROR:
				// CURRENTLY UNUSED
				break;
		}
		
		// Determine if a new state should be entered.
		if (sc.state == SCE_HYPER_DEFAULT) {
			if (IsADigit(sc.ch)) {
				sc.SetState(SCE_HYPER_NUMBER);
			}
			else if (isAWordStart(sc.ch)) {
				sc.SetState(SCE_HYPER_IDENTIFIER);
			}
			else if (visibleChars == 0 && sc.Match('/', '#')) { // begin block comment
				//Platform::DebugPrintf("Start of block comment found\n");
				blockCommentDepth++;
				curLine = styler.GetLine(sc.currentPos);
                styler.SetLineState(curLine, blockCommentDepth);
				sc.SetState(SCE_HYPER_COMMENT_BLOCK);
				sc.Forward();
			}
			else if (visibleChars == 0 && sc.Match('\\', '#')) { // end block
				// this is an invalid end (begin is missing)
				//Platform::DebugPrintf("Loose block stop found (showing error)\n");
				sc.SetState(SCE_HYPER_ERROR);
				sc.Forward();
				sc.Forward();
				sc.SetState(SCE_HYPER_DEFAULT);
			}
			else if (sc.Match('#', '#')) {
				sc.SetState(SCE_HYPER_EOL_COMMENT_LINE);
			}
			else if (sc.Match('#')) { // start of line comment
				sc.SetState(SCE_HYPER_EMB_COMMENT_LINE);
			}
			else if (sc.Match('\"')) {
				sc.SetState(SCE_HYPER_STRING);
			}
			else if (sc.Match('\'')) {
				sc.SetState(SCE_HYPER_CHAR);
			}
			else if (isOperator(sc.ch)) {
				sc.SetState(SCE_HYPER_OPERATOR);
			}
		}
		
		if (!IsASpace(sc.ch) && !IsSpaceEquiv(sc.state)) {
			visibleChars++;
		}
	}
	
	sc.Complete();
}


static void FoldHyperDoc(unsigned int startPos, int length, int,
	WordList *[], Accessor &styler)
{
	
	// not implemented
	
}


static const char * const hyperWordListDesc[] = {
	"Hyper Keywords",
	0
};


LexerModule lmHyper(SCLEX_HYPER, ColouriseHyperDoc, "hyper",
	FoldHyperDoc, hyperWordListDesc);
